require 'uri'
require 'net/http'
require 'json'
require 'forwardable'

module Supai
  class API
    DEFAULT_HOST = ENV.fetch("SUPAI_HOST", "api-un76stg.shop.staging.theunitedfamily.com")

    class Error < Error
      extend Forwardable
      attr_reader :response
      def_delegators :response, :code, :body, :json_body, :uri

      def initialize(message = nil, response:)
        @response = response
        @message = message || [response.code, response.message].join(" ")
        super(@message)
      end
    end

    class Response
      extend Forwardable

      attr_reader :original
      def_delegators :original, :code, :body, :uri, :msg, :message, :inspect, :content_type

      def initialize(obj)
        @original = obj
      end

      def json_body
        @json_body ||= JSON.parse(body)
      end

      def success?
        original.is_a?(Net::HTTPSuccess)
      end

      def error
        API::Error.new(response: self) unless success?
      end
    end

    def initialize(host = DEFAULT_HOST)
      host ||= DEFAULT_HOST # in case we are given nil
      @host = sanitize_host(host)
    end

    def sanitize_host(host)
      host = host.split("://", 2).last
      host = host[0..-2] if host.end_with?("/")
      host
    end

    def request(method, endpoint, body: nil, params: {}, headers: {}, token: nil)
      uri = build_uri(endpoint, params)
      http_start(uri) { |http|
        http.request(
          build_request(
            method,
            uri,
            body: body,
            headers: headers,
            token: token,
          )
        )
      }
    end

  private

    def build_uri(endpoint, params = nil)
      URI.parse("https://#{@host}#{endpoint}").tap { |uri|
        next if params.nil? || params.empty?

        # munge date/times to iso8601 format
        params.each do |k, v|
          if v.respond_to?(:iso8601)
            params[k] = v.iso8601
          end
        end

        uri.query = URI.encode_www_form(params)
      }
    end

    def build_request(method, uri, body: nil, headers: {}, token: nil)
      method_class = case method
                     when :get
                       Net::HTTP::Get
                     when :post
                       Net::HTTP::Post
                     end
      method_class.new(uri.request_uri).tap { |r|
        headers.each do |key, value|
          r[key] = value
        end
        unless token.nil?
          r["Authorization"] = token
        end

        if body.is_a? String
          r.body = body
        elsif !body.nil?
          r["Content-Type"] = "application/json"
          r.body = body.to_json
        end
      }
    end

    def http_start(uri, &block)
      Response.new(Net::HTTP.start(uri.host, uri.port, http_opts, &block))
    end

    def http_opts
      { use_ssl: true }
    end
  end
end
